#******************************************************************************
*                                                                             *
*             An implementation of the p-adic Newton algorithm for            *
*           computing exponential solutions of linear diff. equations         *
*                                                                             *
*                              MuPAD version 1.0                              *
*                                                                             *
*******************************************************************************

Author: Eckhard Pfluegel <Eckhard.Pfluegel@imag.fr>
Last change: 02.05.96 EP

Examples:
=========

eq1 := (93*x^4+9*x^8-51*x^5+42*x^2-33*x^7+38*x^6-78*x^3-36*x+16)*y(x)+(-3*x^5-
9*x^9+12*x^6+6*x^8-2*x^7-12*x^4)*diff(y(x),x)+(3*x^9-4*x^6-2*x^8+3*x^7)*diff(
diff(y(x),x),x);
>> ode::expsols(eq1, y, x, 2);

                   {    /       1  \                  }
                   {    | 3 x + -- |                  {
                   { exp|        2 |       / 3   1  \ {
                   {    \       x  /       | - - -- | {
                   { ---------------, x exp| x    2 | {
                   {        x              \     x  / }


eq2 := (59-16*x^2+8*x)*y(x)+(32*x^2-48*x-110)*diff(y(x),x)+(-16*x^2+64*x^3-52*
       x-59)*diff(diff(y(x),x),x)+(-32*x^2+110+48*x)*diff(diff(diff(y(x),x),x),
       x)+(32*x^2-64*x^3+44*x)*diff(diff(diff(diff(y(x),x),x),x),x);
>> ode::expsols(eq2, y, x, 4);

                             {exp(x), exp(-x)}

eq3 := (23+2*x-2*x^2-24*x^3-18*x^5+122*x^4-8*x^7+7*x^8+90*x^6)*y(x)+(25-9*x+42
       *x^5+2*x^3-338*x^4-68*x^2+48*x^7-396*x^6+6*x^11-24*x^10-175*x^8+23*x^9)*
       diff(y(x),x)+(9*x^2+2-23*x+26*x^6+19*x^4+24*x^8-134*x^7-166*x^5-99*x^3+
       3*x^12-7*x^11-51*x^9+13*x^10)*diff(diff(y(x),x),x);
>> ode::expsols(eq3, y, x, 2);

               {              /   1    \                  }
               {              | ------ |                  {
               { (3 x - 5) exp|  2     |     /     x    \ {
               {              \ x  + 1 /     | - ------ | {
               { -----------------------, exp|    2     | {
               {           2                 \   x  + 1 / {
               {        3 x  + 3                          }

  ode::expsols(eq, y, x, n)
    INPUT:  eq - a linear differential equation with polynomial coefficients
                 or a list of coefficients (for internal use)
            y  - a name (the unknown)
            x  - a name
            n  - an integer (the order of eq)
    OUTPUT: the set of all solutions y of eq with y'/y \in \bar(Q)(x)
#
# 
    This is a first incomplete version computing all solutions
    
    y with y'/y \in Q(x) 
# 
ode::expsols := proc(eq, y, x, n)
local k, lode, l, m, i, j, d, p, parts, parts1, part, places, res, u, f, P;
begin
    if testargs() then
       if args(0)=2 then # case expsols(eq,y(x)) #
          x:=op(y); y:=op(y,0); n:=ode::order(eq,{y});
       end_if
    end_if;
    userinfo(1,"computing exponential solutions");
    if type(eq)=DOM_LIST then lode:=eq
    else
       lode := [];
       for k from 1 to n+1 do 
	   lode := append(lode, coeff(eq,diff(y(x),x$(k-1)),1)) 
       end_for;
    end_if;
    l:=factor(lode[n+1]);
    m := nops(l) div 2;
    # Compute potential singularities #
    for i from 1 to m do p[i]:=l[2*i] end_for; 
    p[m+1] := 1/x;
    parts := []; places := [];
    ode::polysols; # to load ode::padicNewton #
    for i from 1 to m+1 do
        userinfo(2, "computing p-adic exponential parts at: ", p[i]);
        # Compute p-adic exponential parts #
        part := ode::padicNewton(lode, x, p[i], n);
         # Eliminate apparent singularities #
        if part <> [[1, 0, 0]] then
            parts := append(parts, part);
            places := append(places, p[i])
        end_if;
        userinfo(3, "found: ", part);
        if part = [] then 
            userinfo(1, "proof for no exponential solutions");
            return({})
        end_if; 
    end_for;
    # compute possible denominator of rational part #
    userinfo(3, "computing rational part");
    d := 1;
    for i from 1 to nops(parts) do
        for j from 1 to nops(parts[i]) do
            d := lcm(d, parts[i][j][1]);
            parts[i][j] := [parts[i][j][2], parts[i][j][3]]
        end_for
    end_for;
    # and bound it #
    userinfo(3, "bounding denominator:", d);    
    lode := ode::changeRiccati(lode, x, -diff(d, x)/d);
    # eliminate equivalent exponential parts #
    parts1 := [];
    j := 0;
    f := fun([args(1)[2], args(1)[3]]);
    for i from 1 to nops(parts) do
        if parts[i] <> [[0, 0]] then 
            parts1 := append(parts1, parts[i]) 
        end_if
    end_for;
    parts := parts1;
    # combine all p-adic parts #
    parts := ode::possiblePadicParts(parts);
    res := {};
    # test for polynomial solutions #
    for part in parts do
         u := diff(part[1], x)+part[2];
         userinfo(3, "checking Riccati solution:", u);
         for P in ode::polysols(ode::changeRiccati(lode, x, u), x) do
             userinfo(3, "polynomial solution found:", P);
             res := res union {normal(P/d)*exp(part[1]+int(part[2], x))};
         end_for;
         if nops(res) = n then break end_if;
    end_for;
    userinfo(1,"exponential solutions are",op(res));
    res;
end_proc:


# The p-adic Newton algorithm #

# 
  ode::padicNewton(lode, x, p, m)                                             
    INPUT:  lode - list of coefficients of a linear differential operator    
            x    - a name                                                     
            p    - an irreducible polynomial or 1/x (representing infinity)   
            m    - an integer                                                 
    OUTPUT: the list of exponential parts of L at p - at most m parts         
#                                                                            
ode::padicNewton := proc(lode, x, p, m)
local u, m1, res, polygonvar, c, slope;
begin
    res := [];
    userinfo(5, "computing Newton polygon");
    polygonvar := ode::newtonPolygon(lode, x, p, u, m)[3];
    m1 := m;
    while m1 > 0 do
        slope := polygonvar[1][1];
        c := polygonvar[1][2];
        if slope = 0 then
            userinfo(5, "slope = ", slope,
                        "indicial equation =", c);
            if p <> 1/x then
                res := res . ode::padicSingParts(c, x, u, p)
            else
                res := [[1, 0, 0]] 
            end_if;
            m1 := m1  - degree(c, [u]);
        else
            if type(slope) = DOM_INT then
                userinfo(5, "slope = ", slope,
                            "characteristic equation =", c);
                res := res . ode::padicExpParts(lode, c, x, u, p, slope)
            end_if;
            m1 := m1 - degree(c, [u])*denom(slope)
        end_if;
        polygonvar[1] := NIL
    end_while;
    res
end_proc:

# ode::padicSingParts(c, x, u, p) 
    INPUT:  c - a bivariate polynomial in x and u    
            x - a name                   
            u - a name                                                        
            p - an irreducible polynomial in x (or 1/x, representing infinity)
    OUTPUT: the list of p-adic singular parts associated to the generalized   
            exponents ui of the indicial equation c(ui,x) = 0 mod p(x)        
            only the representant mod Z are returned                          
#                                                                             
ode::padicSingParts := proc(c, x, u, p)
    local res, numers, num, i, j, QPhi, P, R, Phi, num1, rest, tr;
begin
    res := [];
    numers := ode::modSolve(c, x, u, p);
    for i from 1 to nops(numers) div 2 do numers[i+1] := NIL end_for;
    numers := ode::reps(numers);
    for i from 1 to nops(numers) do 
        num := numers[i];
        if (type(num) = DOM_INT) then
            if (num < 0) then 
                res := append(res, [p^(-num), 0, 0]) 
            else
                res := append(res, [1, 0, 0])
            end_if
        else
            # We now interprete num as algebraic number: replace in num the   #
            # variable x by Phi where p(Phi) = 0. We then have to compute the #
            # algebraic trace i.e. the sum of all conjugates of num/(x-Phi)., #
            # #
            QPhi := Dom::AlgebraicExtension(Dom::Rational, subs(p, x=Phi));
            P := Dom::DistributedPolynomial([x], QPhi);
            R := Dom::Fraction(P);
            # We divide p by x-Phi in QPhi #
            rest := R::numer(R::normalize(P(p), P(x-Phi)));
            # We then make the denominator rational by multiplying with rest #    
            num1 := P::_mult(P(subs(num, x=Phi)), rest);
            # Now we compute the algebraic trace of all coefficients of num1 #
            tr := 0;
            for j from 0 to degree(p, [x])-1 do
                tr := tr + x^j*ode::conjTrace(P::coeff(num1, x, j), p, x)
            end_for;
            res := append(res, [1, 0, tr/p])
            # This is the corresponding Maple command #
            #res := res,[1, 0, sum(subs(x = a, num[1])/(x-a), a = RootOf(p))];#
        end_if
    end_for;
    if res <> [] then
        userinfo(5, "p-adic singular parts found: ", res)
    end_if;
    res
end_proc:

# ode::reps(lambda)                                                           
    INPUT:  lambda - a list of pairs [u, m] where u is a p-adic exponent     
                     and m its multiplicity                                  
    OUTPUT: a sublist l of lambda such if [u, m] in l, then there is         
            no positive integer i with [u+i, m] in l.                        
#                                                                            
ode::reps := proc(lambda)
    local parts, tmp, p, p1, d, reps;
begin
    reps := {};
    tmp := {op(lambda)};
    parts := tmp;
    while parts <> {} do
        p := op(parts, 1);
        for p1 in parts do
            d := p-p1;
            if type(d) = DOM_INT then
                tmp := tmp minus {p1};
                if d > 0 then p := p1 end_if
            end_if;
        end_for;
        reps := reps union {p};
        parts := tmp
    end_while;
    [op(reps)]
end_proc:

# This is an implementation of the algebraic trace. However, it would be     
  better to include it into the domain "AlgebraicExtension".             # 

#                                                                            
  ode::conjTrace(alpha, P, x)                                                 
    INPUT:  alpha - an element of an algebraic extension over the rationals  
            P     - the minimal polynomial of the extension                  
            x     - the variable of P                                        
    OUTPUT: its algebraic trace, i.e. the sum of the conjugates              
#                                                                            
ode::conjTrace := proc(alpha, P, x)
    local tmp, a, phivar, i, tr; 
begin
    tmp := extop(alpha, 1);
    a := op(tmp, 1); 
    phivar := op(tmp, 2)[1];
    a := subs(a, phivar = x);
    tr := coeff(a, x, 0);
    for i from 1 to degree(P)-1 do
        a := divide(a*x, P, [x], Rem);
        tr := tr + coeff(a, x, i)
    end_for;
    tr
end_proc:
    
# ode::padicExpParts(c, x, u, p, k)                                           
   INPUT:  c - a bivariate polynomial in x and u                             
           x - a name                                                        
           u - a name                                                        
           p - an irreducible polynomial in x or 1/x (representing infinity)
           k - an integer                                                    
   OUTPUT: the list of p-adic exponential parts associated to the slope k and
           the p-adic characteristic equation c(u, x)                        
#                                                                             
ode::padicExpParts := proc(lode, c, x, u, p, k)
    local i, c1, numers, part, rparts, res;
begin
    res := [];
    c1 := subs(c, u = -u*k);
    numers := ode::modSolve(c1, x, u, p);
    for i from 1 to nops(numers) div 2 do
        part := numers[2*i-1]/p^k;
        rparts := ode::padicNewton(
                     ode::changeRiccati(lode, x, diff(part, x)), x, 
                                   p, numers[2*i]);
        res := res . ode::insert(part, rparts);
    end_for;
    res
end_proc:

# ode::insert(part, l)                                                        
   INPUT:  part - an expression                                              
           l    - a list of expressions                                      
   OUTPUT: if l is the empty list then [1, part, 0]                          
           else replace each element e of the form [*, e', *] in l           
           by [*, part+e', *].                                               
#                                                                             
ode::insert := proc(part, l)
    local f;
begin
    if l = [] then [[1, part, 0]] else
        f := fun([args(1)[1], args(1)[2]+args(2), args(1)[3]]);
        map(l, f, part)
    end_if
end_proc:

# ode::possiblePadicParts
    INPUT:  parts - a list of p-adic exponential parts
    OUTPUT: all possible sums of elements of the different exponential 
            parts
#
ode::possiblePadicParts := proc(part) 
    local p1, p2, pp, parts, res;
begin
    if part = [] then return([[0, 0]]) end_if;
    pp := part[1];
    part[1] := NIL;
    parts := ode::possiblePadicParts(part);
    res := [];
    for p1 in pp do
        for p2 in parts do
            res := append(res, [p1[1]+p2[1], p1[2]+p2[2]])
        end_for
    end_for;
    res
end_proc:

# Additive change of riccati solution #

# ode::changeRiccati(L, x, u)                                                 
   INPUT:  L - list of coefficients of linear differential operator          
               with derivation d/dx                                          
           x - a name                                                        
           u - an expression                                                 
   OUTPUT: list of coefficients of linear differential equation which        
           results from the change of unknown variable y(x) to               
           exp(int(u))*z(x).                                                 
#
ode::changeRiccati := proc(L, x, u)
    local n, i, j, T, cs, d;
begin
    n := nops(L);
    T := array(1..n+1);
    T[1] := 1;
    for i from 2 to n+1 do
        T[i] := (normal(diff(T[i-1], x)+u*T[i-1])) 
    end_for;
    cs := array(1..n);
    d := 1;
    for j from 1 to n do
        cs[j] := 0;
        for i from 1 to n-j+1 do
            cs[j] := cs[j] + (binomial(j+i-2, j-1)*T[i]*L[j+i-1])
        end_for;
        cs[j] := normal(cs[j]);
        d := lcm(d, denom(cs[j]))
    end_for;
    for j from 1 to n do cs[j] := normal(cs[j]*d) end_for;
    cs
end_proc:
       
# Solving equations modulo an irreducible polynomial #

# ode::modSolve(c, x, u, p)                                                   
  INPUT:  c - a multivariate polynomial 
          x - a name                                                         
          u - a name                                                         
          p - an irreducible polynomial (or 1/x, representing infinity)      
  OUTPUT: a list of all solutions u \in C[x], deg(u) < deg(p) of c(x, u) = 0 
          (mod p) and their multiplicities                                   
#                                                                            
ode::modSolve := proc(c, x, u, p) 
    local res, d, f, i, s, Qphi, P, C, Phi;
begin
    if p = 1/x then d := 1 else d := degree(p, [x]) end_if;
    res := [];
    if d = 1 then
        f := factor(poly(c, [u]));
        for i from 1 to (nops(f)-1) div 2 do
            s := f[2*i];
            if degree(s, [u]) > 1 then
                print("Warning: Algebraic extension occurred - incomplete solutions");
                res := append(res, 0, f[2*i+1])
            else
                s := op(solve(s, u));
                res := append(res, s, f[2*i+1]);
                userinfo(5, "solution", s);
            end_if
        end_for
    else
        Qphi:=Dom::AlgebraicExtension(Dom::Rational, subs(p, x=Phi));
        P:=Dom::DistributedPolynomial([u],Qphi);
        C := P(subs(c, x=Phi));
        f := factor(C);
        for i from 1 to (nops(f)-1) div 2 do
            s := f[2*i];
            if P::degree(s, u) > 1 then
                print("Warning: Algebraic extension occurred - incomplete solutions");
                res := append(res, 0, f[2*i+1])
            else
                s := subs(op(s, 1), Phi=x);
                s := op(solve(s, u));
                res := append(res, s, f[2*i+1]);
                userinfo(5, "solution", s)
            end_if
        end_for
    end_if;
    res
end_proc:

